/*
* chsums/6q0gA-static.cpp
*
* This file is part of NJet library
* Copyright (C) 2011, 2012 NJet Collaboration
*
* This software is distributed under the terms of the GNU General Public License (GPL)
*/

#include "6q0gA.h"

// class Amp6q0gAAStatic

/* We have 3 abstract flavours, and two types of photons
 * A1 always couples to the first flavour and A2 to the second
 *
 * flav[FC][NN] array, where real flavours correspond to positions,
 * allows abstract flavours to take roles of physical flavours,
 * by changing abstract flavours positions
 *
 * In case of 6q0gAA we need at least 9 such contributions:
 * 3 x (1 with same-line photons + 2 with different-line photons)
 *
 * for instance, for process uxucxcdxd and array {-3,3,-1,1,-2,2}
 * first flavour -1,1 plays the role of cxc
 * second flavour -2,2 plays the role of dxd
 * third flavour -3,3 plays the role of uxu
 * which means that this part has to have:
 * for same-line photons A1u and A1u (because first flavour is Up)
 * for different-line photons A1u and A2d (because second flavour is Down)
 *
 * in fperm[FC][NN] position corresponds to an abstract flavour
 * and value is that flavour's position in fval[FC][NN]
 */

const int
Amp6q0gAAStatic::flav[FC][NN] = {
  // 1 2 3
  {-1,1,-2,2,-3,3},
  {-1,1,-2,2,-3,3},
  {-1,1,-3,3,-2,2},
  {-2,2,-1,1,-3,3},
  {-2,2,-1,1,-3,3},
  {-3,3,-1,1,-2,2},
  {-2,2,-3,3,-1,1},
  {-2,2,-3,3,-1,1},
  {-3,3,-2,2,-1,1},
  // 2 1 3
  {-1,2,-2,1,-3,3},
  {-1,2,-2,1,-3,3},
  {-1,3,-3,1,-2,2},
  {-2,1,-1,2,-3,3},
  {-2,1,-1,2,-3,3},
  {-3,1,-1,3,-2,2},
  {-2,3,-3,2,-1,1},
  {-2,3,-3,2,-1,1},
  {-3,2,-2,3,-1,1},
  // 2 3 1
  {-1,2,-2,3,-3,1},
  {-1,2,-2,3,-3,1},
  {-1,3,-3,2,-2,1},
  {-2,1,-1,3,-3,2},
  {-2,1,-1,3,-3,2},
  {-3,1,-1,2,-2,3},
  {-2,3,-3,1,-1,2},
  {-2,3,-3,1,-1,2},
  {-3,2,-2,1,-1,3},
  // 3 2 1
  {-1,3,-2,2,-3,1},
  {-1,3,-2,2,-3,1},
  {-1,2,-3,3,-2,1},
  {-2,3,-1,1,-3,2},
  {-2,3,-1,1,-3,2},
  {-3,2,-1,1,-2,3},
  {-2,1,-3,3,-1,2},
  {-2,1,-3,3,-1,2},
  {-3,1,-2,2,-1,3},
  // 3 1 2
  {-1,3,-2,1,-3,2},
  {-1,3,-2,1,-3,2},
  {-1,2,-3,1,-2,3},
  {-2,3,-1,2,-3,1},
  {-2,3,-1,2,-3,1},
  {-3,2,-1,3,-2,1},
  {-2,1,-3,2,-1,3},
  {-2,1,-3,2,-1,3},
  {-3,1,-2,3,-1,2},
  // 1 3 2
  {-1,1,-2,3,-3,2},
  {-1,1,-2,3,-3,2},
  {-1,1,-3,2,-2,3},
  {-2,2,-1,3,-3,1},
  {-2,2,-1,3,-3,1},
  {-3,3,-1,2,-2,1},
  {-2,2,-3,1,-1,3},
  {-2,2,-3,1,-1,3},
  {-3,3,-2,1,-1,2},
};

const int
Amp6q0gAAStatic::fvsign[FC] = {
   1, 1, 1, 1, 1, 1, 1, 1, 1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,
   1, 1, 1, 1, 1, 1, 1, 1, 1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,
   1, 1, 1, 1, 1, 1, 1, 1, 1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,
};

const int
Amp6q0gAAStatic::fperm[FC][NN] = {
  {0,1,2,3,4,5},
  {0,1,2,3,4,5},
  {0,1,4,5,2,3},
  {2,3,0,1,4,5},
  {2,3,0,1,4,5},
  {2,3,4,5,0,1},
  {4,5,0,1,2,3},
  {4,5,0,1,2,3},
  {4,5,2,3,0,1},
  //
  {0,3,2,1,4,5},
  {0,3,2,1,4,5},
  {0,3,4,5,2,1},
  {2,1,0,3,4,5},
  {2,1,0,3,4,5},
  {2,1,4,5,0,3},
  {4,5,0,3,2,1},
  {4,5,0,3,2,1},
  {4,5,2,1,0,3},
  //
  {0,5,2,1,4,3},
  {0,5,2,1,4,3},
  {0,5,4,3,2,1},
  {2,1,0,5,4,3},
  {2,1,0,5,4,3},
  {2,1,4,3,0,5},
  {4,3,0,5,2,1},
  {4,3,0,5,2,1},
  {4,3,2,1,0,5},
  //
  {0,5,2,3,4,1},
  {0,5,2,3,4,1},
  {0,5,4,1,2,3},
  {2,3,0,5,4,1},
  {2,3,0,5,4,1},
  {2,3,4,1,0,5},
  {4,1,0,5,2,3},
  {4,1,0,5,2,3},
  {4,1,2,3,0,5},
  //
  {0,3,2,5,4,1},
  {0,3,2,5,4,1},
  {0,3,4,1,2,5},
  {2,5,0,3,4,1},
  {2,5,0,3,4,1},
  {2,5,4,1,0,3},
  {4,1,0,3,2,5},
  {4,1,0,3,2,5},
  {4,1,2,5,0,3},
  //
  {0,1,2,5,4,3},
  {0,1,2,5,4,3},
  {0,1,4,3,2,5},
  {2,5,0,1,4,3},
  {2,5,0,1,4,3},
  {2,5,4,3,0,1},
  {4,3,0,1,2,5},
  {4,3,0,1,2,5},
  {4,3,2,5,0,1},
};

const int
Amp6q0gAAStatic::fvcol[FC][CC] = {
  {0,1,2,3,4,5},
  {0,1,2,3,4,5},
  {1,0,2,4,3,5},
  {1,0,4,3,2,5},
  {1,0,4,3,2,5},
  {0,1,4,2,3,5},
  {0,1,3,4,2,5},
  {0,1,3,4,2,5},
  {1,0,3,2,4,5},
  //
  {4,2,1,5,0,3},
  {4,2,1,5,0,3},
  {2,4,1,0,5,3},
  {2,4,0,5,1,3},
  {2,4,0,5,1,3},
  {4,2,0,1,5,3},
  {4,2,5,0,1,3},
  {4,2,5,0,1,3},
  {2,4,5,1,0,3},
  //
  {1,5,4,2,3,0},
  {1,5,4,2,3,0},
  {5,1,4,3,2,0},
  {5,1,3,2,4,0},
  {5,1,3,2,4,0},
  {1,5,3,4,2,0},
  {1,5,2,3,4,0},
  {1,5,2,3,4,0},
  {5,1,2,4,3,0},
  //
  {2,3,0,1,5,4},
  {2,3,0,1,5,4},
  {3,2,0,5,1,4},
  {3,2,5,1,0,4},
  {3,2,5,1,0,4},
  {2,3,5,0,1,4},
  {2,3,1,5,0,4},
  {2,3,1,5,0,4},
  {3,2,1,0,5,4},
  //
  {5,0,3,4,2,1},
  {5,0,3,4,2,1},
  {0,5,3,2,4,1},
  {0,5,2,4,3,1},
  {0,5,2,4,3,1},
  {5,0,2,3,4,1},
  {5,0,4,2,3,1},
  {5,0,4,2,3,1},
  {0,5,4,3,2,1},
  //
  {3,4,5,0,1,2},
  {3,4,5,0,1,2},
  {4,3,5,1,0,2},
  {4,3,1,0,5,2},
  {4,3,1,0,5,2},
  {3,4,1,5,0,2},
  {3,4,0,1,5,2},
  {3,4,0,1,5,2},
  {4,3,0,5,1,2},
};

const int
Amp6q0gAAStatic::HSarr[HS][HSNN] = {
  { 1,-1, 1,-1, 1,-1, 1, 1},
  { 1,-1, 1,-1,-1, 1, 1, 1},
  { 1,-1,-1, 1, 1,-1, 1, 1},
  { 1,-1,-1, 1,-1, 1, 1, 1},
  {-1, 1, 1,-1, 1,-1, 1, 1},
  {-1, 1, 1,-1,-1, 1, 1, 1},
  {-1, 1,-1, 1, 1,-1, 1, 1},
  {-1, 1,-1, 1,-1, 1, 1, 1},
  { 1,-1, 1,-1, 1,-1,-1, 1},
  { 1,-1, 1,-1,-1, 1,-1, 1},
  { 1,-1,-1, 1, 1,-1,-1, 1},
  { 1,-1,-1, 1,-1, 1,-1, 1},
  {-1, 1, 1,-1, 1,-1,-1, 1},
  {-1, 1, 1,-1,-1, 1,-1, 1},
  {-1, 1,-1, 1, 1,-1,-1, 1},
  {-1, 1,-1, 1,-1, 1,-1, 1},
  { 1,-1, 1,-1, 1,-1, 1,-1},
  { 1,-1, 1,-1,-1, 1, 1,-1},
  { 1,-1,-1, 1, 1,-1, 1,-1},
  { 1,-1,-1, 1,-1, 1, 1,-1},
  {-1, 1, 1,-1, 1,-1, 1,-1},
  {-1, 1, 1,-1,-1, 1, 1,-1},
  {-1, 1,-1, 1, 1,-1, 1,-1},
  {-1, 1,-1, 1,-1, 1, 1,-1},
  { 1,-1, 1,-1, 1,-1,-1,-1},
  { 1,-1, 1,-1,-1, 1,-1,-1},
  { 1,-1,-1, 1, 1,-1,-1,-1},
  { 1,-1,-1, 1,-1, 1,-1,-1},
  {-1, 1, 1,-1, 1,-1,-1,-1},
  {-1, 1, 1,-1,-1, 1,-1,-1},
  {-1, 1,-1, 1, 1,-1,-1,-1},
  {-1, 1,-1, 1,-1, 1,-1,-1},
};

const int
Amp6q0gAA2Static::HSarr[HS][HSNN] = {
  { 1,-1,-1, 1,-1, 1,-1, 1},
  { 1,-1,-1, 1,-1, 1, 1, 1},
  { 1,-1,-1, 1, 1,-1,-1, 1},
  { 1,-1,-1, 1, 1,-1, 1, 1},
  { 1,-1, 1,-1,-1, 1,-1, 1},
  { 1,-1, 1,-1,-1, 1, 1, 1},
  { 1,-1, 1,-1, 1,-1,-1, 1},
  { 1,-1, 1,-1, 1,-1, 1, 1},
  { 1, 1,-1,-1,-1, 1,-1, 1},
  { 1, 1,-1,-1,-1, 1, 1, 1},
  { 1, 1,-1,-1, 1,-1,-1, 1},
  { 1, 1,-1,-1, 1,-1, 1, 1},
  {-1,-1, 1, 1,-1, 1,-1, 1},
  {-1,-1, 1, 1,-1, 1, 1, 1},
  {-1,-1, 1, 1, 1,-1,-1, 1},
  {-1,-1, 1, 1, 1,-1, 1, 1},
  {-1, 1,-1, 1,-1, 1,-1, 1},
  {-1, 1,-1, 1,-1, 1, 1, 1},
  {-1, 1,-1, 1, 1,-1,-1, 1},
  {-1, 1,-1, 1, 1,-1, 1, 1},
  {-1, 1, 1,-1,-1, 1,-1, 1},
  {-1, 1, 1,-1,-1, 1, 1, 1},
  {-1, 1, 1,-1, 1,-1,-1, 1},
  {-1, 1, 1,-1, 1,-1, 1, 1},
  { 1,-1,-1, 1,-1, 1,-1,-1},
  { 1,-1,-1, 1,-1, 1, 1,-1},
  { 1,-1,-1, 1, 1,-1,-1,-1},
  { 1,-1,-1, 1, 1,-1, 1,-1},
  { 1,-1, 1,-1,-1, 1,-1,-1},
  { 1,-1, 1,-1,-1, 1, 1,-1},
  { 1,-1, 1,-1, 1,-1,-1,-1},
  { 1,-1, 1,-1, 1,-1, 1,-1},
  { 1, 1,-1,-1,-1, 1,-1,-1},
  { 1, 1,-1,-1,-1, 1, 1,-1},
  { 1, 1,-1,-1, 1,-1,-1,-1},
  { 1, 1,-1,-1, 1,-1, 1,-1},
  {-1,-1, 1, 1,-1, 1,-1,-1},
  {-1,-1, 1, 1,-1, 1, 1,-1},
  {-1,-1, 1, 1, 1,-1,-1,-1},
  {-1,-1, 1, 1, 1,-1, 1,-1},
  {-1, 1,-1, 1,-1, 1,-1,-1},
  {-1, 1,-1, 1,-1, 1, 1,-1},
  {-1, 1,-1, 1, 1,-1,-1,-1},
  {-1, 1,-1, 1, 1,-1, 1,-1},
  {-1, 1, 1,-1,-1, 1,-1,-1},
  {-1, 1, 1,-1,-1, 1, 1,-1},
  {-1, 1, 1,-1, 1,-1,-1,-1},
  {-1, 1, 1,-1, 1,-1, 1,-1},
};

const int
Amp6q0gAA6Static::HSarr[HS][HSNN] = {
  { 1,-1,-1,-1, 1, 1,-1, 1},
  { 1,-1,-1,-1, 1, 1, 1, 1},
  { 1,-1,-1, 1,-1, 1,-1, 1},
  { 1,-1,-1, 1,-1, 1, 1, 1},
  { 1,-1,-1, 1, 1,-1,-1, 1},
  { 1,-1,-1, 1, 1,-1, 1, 1},
  { 1,-1, 1,-1,-1, 1,-1, 1},
  { 1,-1, 1,-1,-1, 1, 1, 1},
  { 1,-1, 1,-1, 1,-1,-1, 1},
  { 1,-1, 1,-1, 1,-1, 1, 1},
  { 1,-1, 1, 1,-1,-1,-1, 1},
  { 1,-1, 1, 1,-1,-1, 1, 1},
  { 1, 1,-1,-1,-1, 1,-1, 1},
  { 1, 1,-1,-1,-1, 1, 1, 1},
  { 1, 1,-1,-1, 1,-1,-1, 1},
  { 1, 1,-1,-1, 1,-1, 1, 1},
  { 1, 1,-1, 1,-1,-1,-1, 1},
  { 1, 1,-1, 1,-1,-1, 1, 1},
  { 1, 1, 1,-1,-1,-1,-1, 1},
  { 1, 1, 1,-1,-1,-1, 1, 1},
  {-1,-1,-1, 1, 1, 1,-1, 1},
  {-1,-1,-1, 1, 1, 1, 1, 1},
  {-1,-1, 1,-1, 1, 1,-1, 1},
  {-1,-1, 1,-1, 1, 1, 1, 1},
  {-1,-1, 1, 1,-1, 1,-1, 1},
  {-1,-1, 1, 1,-1, 1, 1, 1},
  {-1,-1, 1, 1, 1,-1,-1, 1},
  {-1,-1, 1, 1, 1,-1, 1, 1},
  {-1, 1,-1,-1, 1, 1,-1, 1},
  {-1, 1,-1,-1, 1, 1, 1, 1},
  {-1, 1,-1, 1,-1, 1,-1, 1},
  {-1, 1,-1, 1,-1, 1, 1, 1},
  {-1, 1,-1, 1, 1,-1,-1, 1},
  {-1, 1,-1, 1, 1,-1, 1, 1},
  {-1, 1, 1,-1,-1, 1,-1, 1},
  {-1, 1, 1,-1,-1, 1, 1, 1},
  {-1, 1, 1,-1, 1,-1,-1, 1},
  {-1, 1, 1,-1, 1,-1, 1, 1},
  {-1, 1, 1, 1,-1,-1,-1, 1},
  {-1, 1, 1, 1,-1,-1, 1, 1},
  { 1,-1,-1,-1, 1, 1,-1,-1},
  { 1,-1,-1,-1, 1, 1, 1,-1},
  { 1,-1,-1, 1,-1, 1,-1,-1},
  { 1,-1,-1, 1,-1, 1, 1,-1},
  { 1,-1,-1, 1, 1,-1,-1,-1},
  { 1,-1,-1, 1, 1,-1, 1,-1},
  { 1,-1, 1,-1,-1, 1,-1,-1},
  { 1,-1, 1,-1,-1, 1, 1,-1},
  { 1,-1, 1,-1, 1,-1,-1,-1},
  { 1,-1, 1,-1, 1,-1, 1,-1},
  { 1,-1, 1, 1,-1,-1,-1,-1},
  { 1,-1, 1, 1,-1,-1, 1,-1},
  { 1, 1,-1,-1,-1, 1,-1,-1},
  { 1, 1,-1,-1,-1, 1, 1,-1},
  { 1, 1,-1,-1, 1,-1,-1,-1},
  { 1, 1,-1,-1, 1,-1, 1,-1},
  { 1, 1,-1, 1,-1,-1,-1,-1},
  { 1, 1,-1, 1,-1,-1, 1,-1},
  { 1, 1, 1,-1,-1,-1,-1,-1},
  { 1, 1, 1,-1,-1,-1, 1,-1},
  {-1,-1,-1, 1, 1, 1,-1,-1},
  {-1,-1,-1, 1, 1, 1, 1,-1},
  {-1,-1, 1,-1, 1, 1,-1,-1},
  {-1,-1, 1,-1, 1, 1, 1,-1},
  {-1,-1, 1, 1,-1, 1,-1,-1},
  {-1,-1, 1, 1,-1, 1, 1,-1},
  {-1,-1, 1, 1, 1,-1,-1,-1},
  {-1,-1, 1, 1, 1,-1, 1,-1},
  {-1, 1,-1,-1, 1, 1,-1,-1},
  {-1, 1,-1,-1, 1, 1, 1,-1},
  {-1, 1,-1, 1,-1, 1,-1,-1},
  {-1, 1,-1, 1,-1, 1, 1,-1},
  {-1, 1,-1, 1, 1,-1,-1,-1},
  {-1, 1,-1, 1, 1,-1, 1,-1},
  {-1, 1, 1,-1,-1, 1,-1,-1},
  {-1, 1, 1,-1,-1, 1, 1,-1},
  {-1, 1, 1,-1, 1,-1,-1,-1},
  {-1, 1, 1,-1, 1,-1, 1,-1},
  {-1, 1, 1, 1,-1,-1,-1,-1},
  {-1, 1, 1, 1,-1,-1, 1,-1},
};
